#           .---.        .-----------
#          /     \  __  /    ------
#         / /     \(..)/    -----
#        //////   ' \/ `   ---
#       //// / // :    : ---
#      // /   /  /`    '--
#     // /        //..\\
#   o===|========UU====UU=====-  -==========================o
#                '//||\\`
#                       DEVELOPED BY JGH
#
#   -=====================|===o  o===|======================-+

import numpy as np
import pandas as pd


class Stat_Tables:
    """Statistic tables paremeter"""

    def __init__(self):
        """Initializate tables object loading
        tables parameter into a dictionary

        Methods:
        ---------
        get_A : Float
            Return the A constant

        get_A2 : Float
            Return the A2 constant

        get_A3 : Float
            Return the A3 constant

        get_c4 : Float
            Return the c4 constant

        get_B3 : Float
            Return the B3 constant

        get_B4 : Float
            Return the B4 constant

        get_B5 : Float
            Return the B5 constant

        get_B6 : Float
            Return the B6 constant

        get_B4 : Float
            Return the B4 constant

        get_d2 : Float
            Return the d2 constant

        get_d3 : Float
            Return the d3 constant

        get_D1 : Float
            Return the D1 constant

        get_D2 : Float
            Return the D2 constant

        get_D3 : Float
            Return the D3 constant

        get_D4 : Float
            Return the D4 constant
        """
        self.__dict_info = {
            "A": {
                2: 2.121,
                3: 1.732,
                4: 1.5,
                5: 1.342,
                6: 1.225,
                7: 1.134,
                8: 1.061,
                9: 1,
                10: 0.949,
                11: 0.905,
                12: 0.866,
                13: 0.832,
                14: 0.802,
                15: 0.775,
                16: 0.75,
                17: 0.728,
                18: 0.707,
                19: 0.688,
                20: 0.671,
                21: 0.655,
                22: 0.64,
                23: 0.626,
                24: 0.612,
                25: 0.6,
            },
            "A2": {
                2: 1.88,
                3: 1.023,
                4: 0.729,
                5: 0.577,
                6: 0.483,
                7: 0.419,
                8: 0.373,
                9: 0.337,
                10: 0.308,
                11: 0.285,
                12: 0.266,
                13: 0.249,
                14: 0.235,
                15: 0.223,
                16: 0.212,
                17: 0.203,
                18: 0.194,
                19: 0.187,
                20: 0.18,
                21: 0.173,
                22: 0.167,
                23: 0.162,
                24: 0.157,
                25: 0.153,
            },
            "A3": {
                2: 2.659,
                3: 1.954,
                4: 1.628,
                5: 1.427,
                6: 1.287,
                7: 1.182,
                8: 1.099,
                9: 1.032,
                10: 0.975,
                11: 0.927,
                12: 0.886,
                13: 0.85,
                14: 0.817,
                15: 0.789,
                16: 0.763,
                17: 0.739,
                18: 0.718,
                19: 0.698,
                20: 0.68,
                21: 0.663,
                22: 0.647,
                23: 0.633,
                24: 0.619,
                25: 0.606,
            },
            "c4": {
                2: 0.7979,
                3: 0.8862,
                4: 0.9213,
                5: 0.94,
                6: 0.9515,
                7: 0.9594,
                8: 0.965,
                9: 0.9693,
                10: 0.9727,
                11: 0.9754,
                12: 0.9776,
                13: 0.9794,
                14: 0.981,
                15: 0.9823,
                16: 0.9835,
                17: 0.9845,
                18: 0.9854,
                19: 0.9862,
                20: 0.9869,
                21: 0.9876,
                22: 0.9882,
                23: 0.9887,
                24: 0.9892,
                25: 0.9896,
            },
            "B3": {
                2: 0,
                3: 0,
                4: 0,
                5: 0,
                6: 0.03,
                7: 0.118,
                8: 0.185,
                9: 0.239,
                10: 0.284,
                11: 0.321,
                12: 0.354,
                13: 0.382,
                14: 0.406,
                15: 0.428,
                16: 0.448,
                17: 0.466,
                18: 0.482,
                19: 0.497,
                20: 0.51,
                21: 0.523,
                22: 0.534,
                23: 0.545,
                24: 0.555,
                25: 0.565,
            },
            "B4": {
                2: 3.267,
                3: 2.568,
                4: 2.266,
                5: 2.089,
                6: 1.97,
                7: 1.882,
                8: 1.815,
                9: 1.761,
                10: 1.716,
                11: 1.679,
                12: 1.646,
                13: 1.618,
                14: 1.594,
                15: 1.572,
                16: 1.552,
                17: 1.534,
                18: 1.518,
                19: 1.503,
                20: 1.49,
                21: 1.477,
                22: 1.466,
                23: 1.455,
                24: 1.445,
                25: 1.435,
            },
            "B5": {
                2: 0,
                3: 0,
                4: 0,
                5: 0,
                6: 0.029,
                7: 0.113,
                8: 0.179,
                9: 0.232,
                10: 0.276,
                11: 0.313,
                12: 0.346,
                13: 0.374,
                14: 0.399,
                15: 0.421,
                16: 0.44,
                17: 0.458,
                18: 0.475,
                19: 0.49,
                20: 0.504,
                21: 0.516,
                22: 0.528,
                23: 0.539,
                24: 0.549,
                25: 0.559,
            },
            "B6": {
                2: 2.606,
                3: 2.276,
                4: 2.088,
                5: 1.964,
                6: 1.874,
                7: 1.806,
                8: 1.751,
                9: 1.707,
                10: 1.669,
                11: 1.637,
                12: 1.61,
                13: 1.585,
                14: 1.563,
                15: 1.544,
                16: 1.526,
                17: 1.511,
                18: 1.496,
                19: 1.483,
                20: 1.47,
                21: 1.459,
                22: 1.448,
                23: 1.438,
                24: 1.429,
                25: 1.42,
            },
            "d2": {
                2: 1.128,
                3: 1.693,
                4: 2.059,
                5: 2.326,
                6: 2.534,
                7: 2.704,
                8: 2.847,
                9: 2.97,
                10: 3.078,
                11: 3.173,
                12: 3.258,
                13: 3.336,
                14: 3.407,
                15: 3.472,
                16: 3.532,
                17: 3.588,
                18: 3.64,
                19: 3.689,
                20: 3.735,
                21: 3.778,
                22: 3.819,
                23: 3.858,
                24: 3.895,
                25: 3.931,
            },
            "d3": {
                2: 0.852,
                3: 0.888,
                4: 0.879,
                5: 0.864,
                6: 0.848,
                7: 0.833,
                8: 0.819,
                9: 0.807,
                10: 0.797,
                11: 0.787,
                12: 0.778,
                13: 0.77,
                14: 0.763,
                15: 0.756,
                16: 0.75,
                17: 0.744,
                18: 0.738,
                19: 0.733,
                20: 0.728,
                21: 0.724,
                22: 0.719,
                23: 0.715,
                24: 0.712,
                25: 0.708,
            },
            "D1": {
                2: 0,
                3: 0,
                4: 0,
                5: 0,
                6: 0,
                7: 0.206,
                8: 0.389,
                9: 0.548,
                10: 0.688,
                11: 0.813,
                12: 0.924,
                13: 1.026,
                14: 1.119,
                15: 1.204,
                16: 1.283,
                17: 1.357,
                18: 1.425,
                19: 1.49,
                20: 1.55,
                21: 1.607,
                22: 1.661,
                23: 1.712,
                24: 1.761,
                25: 1.807,
            },
            "D2": {
                2: 3.686,
                3: 4.357,
                4: 4.697,
                5: 4.918,
                6: 5.078,
                7: 5.203,
                8: 5.306,
                9: 5.392,
                10: 5.467,
                11: 5.533,
                12: 5.593,
                13: 5.646,
                14: 5.695,
                15: 5.739,
                16: 5.781,
                17: 5.819,
                18: 5.855,
                19: 5.888,
                20: 5.92,
                21: 5.95,
                22: 5.978,
                23: 6.004,
                24: 6.03,
                25: 6.055,
            },
            "D3": {
                2: 0,
                3: 0,
                4: 0,
                5: 0,
                6: 0,
                7: 0.076,
                8: 0.137,
                9: 0.184,
                10: 0.223,
                11: 0.256,
                12: 0.284,
                13: 0.307,
                14: 0.328,
                15: 0.347,
                16: 0.363,
                17: 0.378,
                18: 0.392,
                19: 0.404,
                20: 0.415,
                21: 0.425,
                22: 0.435,
                23: 0.444,
                24: 0.452,
                25: 0.46,
            },
            "D4": {
                2: 3.266,
                3: 2.574,
                4: 2.281,
                5: 2.114,
                6: 2.003,
                7: 1.924,
                8: 1.863,
                9: 1.816,
                10: 1.777,
                11: 1.744,
                12: 1.716,
                13: 1.693,
                14: 1.672,
                15: 1.653,
                16: 1.637,
                17: 1.622,
                18: 1.608,
                19: 1.596,
                20: 1.585,
                21: 1.575,
                22: 1.565,
                23: 1.556,
                24: 1.548,
                25: 1.54,
            },
        }

    def get_A(self, n):
        """Return the A constant

        Args:
        ------
        n : int
            Number of items

        Returns:
        --------
        Float : A constant for the given number of items

        Raises:
        --------
        TypeError

        A_01
            n must be an integer

        A_02
            Number of items must be >=2
        """
        if type(n) is not int:
            raise ValueError(
                f"Error A_01: Items must be an integer. Specified: {n}"
            )
        if n < 2:
            raise ValueError(f"Error A_02: Items must be >=2. Specified: {n}")
        elif n >= 2 and n <= 25:
            return self.__dict_info["A"][n]
        else:
            return round(3 / np.sqrt(n), 4)

    def get_A2(self, n, bol_Pass=False):
        """Return the A2 constant

        Args:
        ------
        n : int
            Number of items

        bol_Pass: Boolean
            False to raise error for n>25, True for A2_25 when n>25

        Returns:
        ---------
        Float : A2 constant for the given number of items

        Raises:
        --------
        TypeError

        A_01
            n must be an integer

        A_02
            Number of items must be >=2

        A_03
            Number of items is higher than 25. Set bol_Pass=True to avoid
            this error and alue for n=25 wil be returned
        """
        if type(n) is not int:
            raise ValueError(
                f"Error A_01: Items number must be an integer. Specified: {n}"
            )
        if n < 2:
            raise ValueError(
                f"Error A_02: Samples must be >=2. Specified: {n}"
            )
        elif n > 25 and bol_Pass is False:
            raise ValueError(
                f"Error A_03: items number is higher than 25. Specified: {n}."
                + " Set bol_Pass=True to avoid this error and alue for n=25"
                + " will be returned"
            )
        elif n > 25 and bol_Pass is True:
            return self.__dict_info["A2"][25]
        else:
            return self.__dict_info["A2"][n]

    def get_c4(self, n):
        """Return the c4 constant

        Args:
        ------
        n : int
            Number of items

        Returns:
        ---------
        Float : c4 constant for the given number of items

        Raises:
        --------
        TypeError

        A_01
            n must be an integer

        A_02
            Number of items must be >=2
        """
        if type(n) is not int:
            raise ValueError(
                f"Error A_01: Items must be an integer. Specified: {n}"
            )
        if n < 2:
            raise ValueError(f"Error A_02: Items must be >=2. Specified: {n}")
        elif n >= 2 and n <= 25:
            return self.__dict_info["c4"][n]
        else:
            return round(4 * (n - 1) / (4 * n - 3), 4)

    def get_A3(self, n):
        """Return the A3 constant

        Args:
        ------
        n : int
            Number of items

        Returns:
        ---------
        Float : A3 constant for the given number of items

        Raises:
        --------
        TypeError

        A_01
            n must be an integer

        A_02
            Number of items must be >=2
        """
        if type(n) is not int:
            raise ValueError(
                f"Error A_01: Items must be an integer. Specified: {n}"
            )
        if n < 2:
            raise ValueError(f"Error A_02: Items must be >=2. Specified: {n}")
        elif n >= 2 and n <= 25:
            return self.__dict_info["A3"][n]
        else:
            return round(3 / (self.get_c4(n) * np.sqrt(n)), 4)

    def get_B3(self, n):
        """Return the B3 constant

        Args:
        ------
        n : int
            Number of items

        Returns:
        ---------
        Float : B3 constant for the given number of items

        Raises:
        --------
        TypeError

        A_01
            n must be an integer

        A_02
            Number of items must be >=2
        """
        if type(n) is not int:
            raise ValueError(
                f"Error A_01: Items must be an integer. Specified: {n}"
            )
        if n < 2:
            raise ValueError(f"Error A_02: Items must be >=2. Specified: {n}")
        elif n >= 2 and n <= 25:
            return self.__dict_info["B3"][n]
        else:
            return round(1 - 3 / (self.get_c4(n) * np.sqrt(2 * (n - 1))), 4)

    def get_B4(self, n):
        """Return the B4 constant

        Args:
        ------
        n : int
            Number of items

        Returns:
        ---------
        Float : B4 constant for the given number of items

        Raises:
        --------
        TypeError

        A_01
            n must be an integer

        A_02
            Number of items must be >=2
        """
        if type(n) is not int:
            raise ValueError(
                f"Error A_01: Items must be an integer. Specified: {n}"
            )
        if n < 2:
            raise ValueError(f"Error A_02: Items must be >=2. Specified: {n}")
        elif n >= 2 and n <= 25:
            return self.__dict_info["B4"][n]
        else:
            return round(1 + 3 / (self.get_c4(n) * np.sqrt(2 * (n - 1))), 4)

    def get_B5(self, n):
        """Return the B5 constant

        Args:
        ------
        n : int
            Number of items

        Returns:
        ---------
        Float : B5 constant for the given number of items

        Raises:
        --------
        TypeError

        A_01
            n must be an integer

        A_02
            Number of items must be >=2
        """
        if type(n) is not int:
            raise ValueError(
                f"Error A_01: Items must be an integer. Specified: {n}"
            )
        if n < 2:
            raise ValueError(f"Error A_02: Items must be >=2. Specified: {n}")
        elif n >= 2 and n <= 25:
            return self.__dict_info["B5"][n]
        else:
            return round(self.get_c4(n) - 3 / np.sqrt(2 * (n - 1)), 4)

    def get_B6(self, n):
        """Return the B6 constant

        Args:
        ------
        n : int
            Number of items

        Returns:
        ---------
        Float : B6 constant for the given number of items

        Raises:
        --------
        TypeError

        A_01
            n must be an integer

        A_02
            Number of items must be >=2
        """
        if type(n) is not int:
            raise ValueError(
                f"Error A_01: Items must be an integer. Specified: {n}"
            )
        if n < 2:
            raise ValueError(f"Error A_02: Items must be >=2. Specified: {n}")
        elif n >= 2 and n <= 25:
            return self.__dict_info["B6"][n]
        else:
            return round(self.get_c4(n) + 3 / np.sqrt(2 * (n - 1)), 4)

    def get_d2(self, n):
        """Return the d2 constant

        Args:
        ------
        n : int
            Number of items

        Returns:
        ---------
        Float : d2 constant for the given number of items

        Raises:
        --------
        TypeError

        A_01
            n must be an integer

        A_02
            Number of items must be >=2
        """
        if type(n) is not int:
            raise ValueError(
                f"Error A_01: Items must be an integer. Specified: {n}"
            )
        if n < 2:
            raise ValueError(f"Error A_02: Items must be >=2. Specified: {n}")
        elif n >= 2 and n <= 25:
            return self.__dict_info["d2"][n]
        else:
            mu, sigma = 0, 1
            x = pd.DataFrame()
            for i in range(1, n + 1):
                x["x" + str(i)] = np.random.normal(mu, sigma, 10000000)
            x["range"] = abs(x.min(axis=1) - x.max(axis=1))
            return round(x["range"].mean(axis=0).round(3), 4)

    def get_d3(self, n, bol_Pass=False):
        """Return the d3 constant

        Args:
        ------
        n : int
            Number of items

        bol_Pass: Boolean
            False to raise error for n>25, True for d3_25 when n>25

        Returns:
        ---------
        Float : d3 constant for the given number of items

        Raises:
        --------
        TypeError

        A_01
            n must be an integer

        A_02
            Number of items must be >=2

        A_03
            Number of items is higher than 25. Set bol_Pass=True to avoid
            this error and alue for n=25 wil be returned
        """
        if type(n) is not int:
            raise ValueError(
                f"Error A_01: Items number must be an integer. Specified: {n}"
            )
        if n < 2:
            raise ValueError(
                f"Error A_02: Samples must be >=2. Specified: {n}"
            )
        elif n > 25 and bol_Pass is False:
            raise ValueError(
                f"Error A_03: Items number is higher than 25. Specified: {n}."
                + " Set bol_Pass=True to avoid this error and alue for n=25"
                + " will be returned"
            )
        elif n > 25 and bol_Pass is True:
            return self.__dict_info["d3"][25]
        else:
            return self.__dict_info["d3"][n]

    def get_D1(self, n, bol_Pass=False):
        """Return the D1 constant

        Args:
        ------
        n : int
            Number of items

        bol_Pass: Boolean
            False to raise error for n>25, True for D1_25 when n>25

        Returns:
        ---------
        Float : D1 constant for the given number of items

        Raises:
        --------
        TypeError

        A_01
            n must be an integer

        A_02
            Number of items must be >=2

        A_03
            Number of items is higher than 25. Set bol_Pass=True to avoid
            this error and alue for n=25 wil be returned
        """
        if type(n) is not int:
            raise ValueError(
                f"Error A_01: Items number must be an integer. Specified: {n}"
            )
        if n < 2:
            raise ValueError(
                f"Error A_02: Samples must be >=2. Specified: {n}"
            )
        elif n > 25 and bol_Pass is False:
            raise ValueError(
                f"Error A_03: Items number is higher than 25. Specified: {n}."
                + " Set bol_Pass=True to avoid this error and alue for n=25"
                + " will be returned"
            )
        elif n > 25 and bol_Pass is True:
            return self.__dict_info["D1"][25]
        else:
            return self.__dict_info["D1"][n]

    def get_D2(self, n, bol_Pass=False):
        """Return the D2 constant

        Args:
        ------
        n : int
            Number of items

        bol_Pass: Boolean
            False to raise error for n>25, True for D2_25 when n>25

        Returns:
        ---------
        Float : D2 constant for the given number of items

        Raises:
        --------
        TypeError

        A_01
            n must be an integer

        A_02
            Number of items must be >=2

        A_03
            Number of items is higher than 25. Set bol_Pass=True to avoid
            this error and alue for n=25 wil be returned
        """
        if type(n) is not int:
            raise ValueError(
                f"Error A_01: Items number must be an integer. Specified: {n}"
            )
        if n < 2:
            raise ValueError(
                f"Error A_02: Samples must be >=2. Specified: {n}"
            )
        elif n > 25 and bol_Pass is False:
            raise ValueError(
                f"Error A_03: Items number is higher than 25. Specified: {n}."
                + " Set bol_Pass=True to avoid this error and alue for n=25"
                + " will be returned"
            )
        elif n > 25 and bol_Pass is True:
            return self.__dict_info["D2"][25]
        else:
            return self.__dict_info["D2"][n]

    def get_D3(self, n, bol_Pass=False):
        """Return the D3 constant

        Args:
        ------
        n : int
            Number of items

        bol_Pass: Boolean
            False to raise error for n>25, True for D3_25 when n>25

        Returns:
        ---------
        Float : D3 constant for the given number of items

        Raises:
        --------
        TypeError

        A_01
            n must be an integer

        A_02
            Number of items must be >=2

        A_03
            Number of items is higher than 25. Set bol_Pass=True to avoid
            this error and alue for n=25 wil be returned
        """
        if type(n) is not int:
            raise ValueError(
                f"Error A_01: Items number must be an integer. Specified: {n}"
            )
        if n < 2:
            raise ValueError(
                f"Error A_02: Samples must be >=2. Specified: {n}"
            )
        elif n > 25 and bol_Pass is False:
            raise ValueError(
                f"Error A_03: Items number is higher than 25. Specified: {n}."
                + " Set bol_Pass=True to avoid this error and alue for n=25"
                + " will be returned"
            )
        elif n > 25 and bol_Pass is True:
            return self.__dict_info["D3"][25]
        else:
            return self.__dict_info["D3"][n]

    def get_D4(self, n, bol_Pass=False):
        """Return the D4 constant

        Args:
        ------
        n : int
            Number of items

        bol_Pass: Boolean
            False to raise error for n>25, True for D4_25 when n>25

        Returns:
        ---------
        Float : D4 constant for the given number of items

        Raises:
        --------
        TypeError

        A_01
            n must be an integer

        A_02
            Number of items must be >=2

        A_03
            Number of items is higher than 25. Set bol_Pass=True to avoid
            this error and alue for n=25 wil be returned
        """
        if type(n) is not int:
            raise ValueError(
                f"Error A_01: Items number must be an integer. Specified: {n}"
            )
        if n < 2:
            raise ValueError(
                f"Error A_02: Samples must be >=2. Specified: {n}"
            )
        elif n > 25 and bol_Pass is False:
            raise ValueError(
                f"Error A_03: Items number is higher than 25. Specified: {n}."
                + " Set bol_Pass=True to avoid this error and alue for n=25"
                + " will be returned"
            )
        elif n > 25 and bol_Pass is True:
            return self.__dict_info["D4"][25]
        else:
            return self.__dict_info["D4"][n]



#           .---.        .-----------
#          /     \  __  /    ------
#         / /     \(..)/    -----
#        //////   ' \/ `   ---
#       //// / // :    : ---
#      // /   /  /`    '--
#     // /        //..\\
#   o===|========UU====UU=====-  -==========================o
#                '//||\\`
#                       DEVELOPED BY JGH
#
#   -=====================|===o  o===|======================-+
import statistics

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pylab
from scipy import stats


class Capability:
    """Capability analysis works as a model.
    Once input parameter are specified, model is solved and available
     to print different reports depending on the user requirement

    Args:
    -----
    mydf : pandas dataframe
      At least the measures in a column. For LT analysis, batch column too.

    mydict : dictionary
      Value --> String. Measures column name | Batch --> Optional String.
      Batch ID column name | LSL --> Optional float. Lower specification
       limit | HSL --> Optional float. Higher specification limit |
       Goal --> Optional float. Goal. ONE OF LSL OR HSL MUST BE SPECIFIED

    Methods:
    ----------
    getLog: string
          printable string containing all individual calculations

    Normality_test: matplotlib figure
          report to check sample normality

    Report: matplotlib figure
          Capability report

    Raises:
    ---------
    TypeError

    Init_01
        mydict keys ar not correctly defined

    Init_02
        Specified measures column name is not found

    Init_03
        Dataframe contains null values

    Init_04
        Specified measures column a non numerical type column
    """

    def __init__(self, mydf, mydict):  # noqa
        """Initializate a new instance of a capability model"""
        # check dict is correctly defined
        lst_key = ["value", "batch", "LSL", "HSL", "goal"]
        if not all(key in mydict for key in lst_key):
            raise ValueError(
                f"Error init_01: wrong arg dictionary keys: {mydict.keys()} |"
                f" Be sure to use: {lst_key}"
            )
        lst_bol = [
            isinstance(mydict["LSL"], int),
            isinstance(mydict["LSL"], float),
            isinstance(mydict["HSL"], int),
            isinstance(mydict["HSL"], float),
        ]

        if True not in lst_bol:
            raise ValueError(
                f"Error init_05: neither LSL or HSL are numeric."
                f"LSL: {mydict['LSL']}"
                f" ,HSL: {mydict['HSL']} |"
                f" One must be numeric"
            )

        lst_goal = [
            isinstance(mydict["goal"], int),
            isinstance(mydict["goal"], float),
        ]
        if True in lst_goal:
            self.__goal = True
        else:
            self.__goal = False

        # Initializate different variables
        self.__log = list()
        df_cp = pd.DataFrame()

        # Save which tolerance is specified
        self.__Tol = 0  # None
        if lst_bol[0] is True or lst_bol[1] is True:
            self.__log.append("LSL is specified")
            self.__Tol = 1  # LSL is specified

        if True in [lst_bol[2], lst_bol[3]] and self.__Tol == 1:
            self.__log.append("HSL is specified")
            self.__Tol = 3  # HSL and LSL are specified
        elif True in [lst_bol[2], lst_bol[3]]:
            self.__log.append("HSL is specified")
            self.__Tol = 2  # HSL is specified

        # check measures data column is correct defined
        if mydict["value"] not in mydf.keys().tolist():
            raise ValueError(
                f"Error init_02: Specified dict_value: {mydict['value']} |"
                f" is not in the original dataframe column names:"
                f" {mydf.keys().tolist()}, please, review the class argument"
            )
        df_cp["Value"] = mydf[mydict["value"]]

        # Check if null values exist
        if df_cp.isna().sum().sum() > 0:
            raise ValueError(
                "Error init_03: Given dataframe contains"
                f" {df_cp.isna().sum().sum()} null values. Please, clean"
                " and filter your dataset before to call this method"
            )

        # Check if measures is a numeric column
        if pd.api.types.is_numeric_dtype(df_cp["Value"]) is False:
            raise ValueError(
                f"Error init_04: Given dataframe column {mydict['value']} must"
                " be numeric type. Please, review your dataset"
            )

        # In docs, remember remark the importance to read docs if long term
        # analysis is not done
        # Check if batch is defined. If so, stablish as categorical data type
        if mydict["batch"] not in mydf.keys().tolist():
            self.__log.append(
                f"{mydict['batch']} not match with column names.so"
                f" {mydf.keys().tolist()}"
            )
            self.__log.append("Short term analysis is activated")
            self.__AnalysisType = 1
        else:
            self.__log.append("Long term analysis is activated")
            self.__AnalysisType = 2
            df_cp["Batch"] = mydf[mydict["batch"]].astype("category")

        self.__df_cp = df_cp
        self.__log.append("Model is created")
        self.__mydict = mydict

    def getLog(self):
        """Return a string which contain all important calculations.

        Returns:
        ---------
        log: String
            all step logged
        """
        # Build up log string to be printed
        str_log = ""
        for i in range(0, len(self.__log)):
            str_log = str_log + self.__log[i] + "\n"

        return str_log

    def Normality_test(self):
        """Normality_test report is a figure that contain different chart and
        data description which helps to conclude if the measurement could be
        explained as a normal distribution so capability analysis could be
        done, or, inthe other hand, this analysis could not be take place.
        First, an histogram is showed, then, the probability plot. Third one
        is a time series plot and a descriptive data box.
        Finally, an advisement is showed based on p value.

        Returns:
        ---------
        Fig_NT : matplotlib figure
            Set of charts
        """
        df_work = self.__df_cp.copy()
        Fig_NT = plt.figure(figsize=(18, 12))
        Fig_NT.set_facecolor("white")
        gs = mpl.gridspec.GridSpec(2, 2, wspace=0.2, hspace=0.2)
        Fig_NT.suptitle("\nNormality Test", fontsize=20)

        # ============================================================================================
        #                                Histogram
        # ============================================================================================
        ax1 = Fig_NT.add_subplot(gs[0, 0])
        ax1.set_title("Histogram")
        myweights = (
            np.ones(len(df_work["Value"])) / len(df_work["Value"]) * 100
        )
        ax1.hist(df_work["Value"], bins=100, weights=myweights)
        ax1.set_ylabel("% Weight")
        ax1.set_xlabel("Variable")

        # ============================================================================================
        #                                Probability Plot
        # ============================================================================================
        ax2 = Fig_NT.add_subplot(gs[0, 1])

        measurement = df_work["Value"].dropna().to_numpy()
        stat, p = stats.shapiro(measurement)
        ax2 = stats.probplot(measurement, dist="norm", plot=pylab)  # noqa

        # ============================================================================================
        #                                Time series
        # ============================================================================================
        ax3 = Fig_NT.add_subplot(gs[1, 0])
        ax3.set_title("Time series")
        ax3.plot(df_work.index, df_work["Value"], marker=".")
        ax3.set_ylabel("Variable")
        ax3.set_xlabel("Time")

        # ============================================================================================
        #                                Descriptive
        # ============================================================================================
        ax4 = Fig_NT.add_subplot(gs[1, 1])
        ax4.set_xticks([])
        ax4.set_yticks([])
        ax4.set_facecolor("white")
        ax4.set_ylim(0, 1)
        ax4.set_xlim(0, 1)
        ax4.annotate(
            "DESCRIPTIVE STATISTIC\n\n{}\n\np value: {:.2f}".format(
                df_work.describe(), p
            ),
            xy=(0.3, 0.2),
            bbox=dict(boxstyle="round", fc="w", color="black"),
            fontsize=12,
        )
        if p > 0.05:
            ax4.annotate(
                "Process seems to match with Normal distribution",
                xy=(0.1, 0.05),
                bbox=dict(boxstyle="round", fc="mediumseagreen"),
                fontsize=12,
                fontweight="bold",
            )
        else:
            ax4.annotate(
                "Process seems to NOT match with Normal distribution",
                xy=(0.05, 0.05),
                bbox=dict(boxstyle="round", fc="red"),
                fontsize=12,
                fontweight="bold",
            )

        return Fig_NT

    def __Cp_Report(self):
        """Cp draws short term report

        Returns:
        ---------
        Fig_Cp : matplotlib figure
            Short Term Report
        """
        # Determine avg and standard deviation
        df_cp = self.__df_cp
        dbl_Total_avg = self.__df_cp["Value"].mean()
        SD = statistics.stdev(df_cp["Value"])

        # Solve Capability parameters
        if self.__Tol == 3:
            Cp = round(
                (self.__mydict["HSL"] - self.__mydict["LSL"]) / (6 * SD), 2
            )
            CpL = round((dbl_Total_avg - self.__mydict["LSL"]) / (3 * SD), 2)
            CpU = round((self.__mydict["HSL"] - dbl_Total_avg) / (3 * SD), 2)
            Cpk = min(CpL, CpU)
        elif self.__Tol == 1:
            Cp = "*"
            CpL = round((dbl_Total_avg - self.__mydict["LSL"]) / (3 * SD), 2)
            CpU = "*"
            Cpk = CpL
        elif self.__Tol == 2:
            Cp = "*"
            CpL = "*"
            CpU = round((self.__mydict["HSL"] - dbl_Total_avg) / (3 * SD), 2)
            Cpk = CpU

        # build up the model
        dbl_step = (df_cp["Value"].max() - df_cp["Value"].min()) * 0.01
        np_values = np.arange(
            df_cp["Value"].min(), df_cp["Value"].max(), dbl_step
        )

        Fig_Cp = plt.figure(figsize=(18, 12))
        Fig_Cp.set_facecolor("white")
        gs = mpl.gridspec.GridSpec(2, 4, wspace=0.2, hspace=0.2)
        Fig_Cp.suptitle(
            "\nProcess capability report for {}".format(
                self.__mydict["value"]
            ),
            fontsize=20,
        )

        # ============================================================================================
        #                                DATA
        # ============================================================================================
        ax1 = Fig_Cp.add_subplot(gs[0, 0])
        ax1.set_xticks([])
        ax1.set_yticks([])
        ax1.set_facecolor("white")
        ax1.set_ylim(0, 1)
        ax1.set_xlim(0, 1)

        str_msg = (
            f"{'':^40s}\n\n"
            f"{'':^35s}\n"
            f"{'':^35s}\n"
            f"{'':^35s}\n"
            f"{'':^35s}\n"
            f"{'':^35s}\n"
            f"{'':^35s}\n"
            f"{'':^35s}\n"
        )

        ax1.text(
            0.25,
            0.21,
            str_msg,
            bbox=dict(boxstyle="round", fc="w", color="black"),
            fontsize=12,
        )
        str_msg = f"{'Data Processing':^25s}"
        ax1.text(0.23, 0.58, str_msg, fontsize=12)

        str_msg = (
            f"{' LSL':15s}\n"
            f"{' Goal':15s}\n"
            f"{' HSL':15s}\n"
            f"{' Mean':15s}\n"
            f"{' Sample Size':15s}\n"
            f"{'':15s}\n"
            f"{' Std. Dev (ST)':15s}\n"
        )
        ax1.text(0.25, 0.2, str_msg, fontsize=12)

        str_msg = (
            f"{self.__mydict['LSL']:<10} \n"
            f"{self.__mydict['goal']:<10} \n"
            f"{self.__mydict['HSL']:<10} \n"
            f"{dbl_Total_avg:<10.3g} \n"
            f"{len(df_cp.index):<10.3g} \n"
            f"{'':<10} \n"
            f"{SD:<10.3g} \n"
        )
        ax1.text(0.66, 0.2, str_msg, fontsize=12)

        # ============================================================================================
        #                                HISTOGRAM
        # ============================================================================================
        ax2 = Fig_Cp.add_subplot(gs[0, 1:3])
        ax2.hist(df_cp["Value"], bins=100, label="Data")
        ax2.set_yticks([])

        if self.__Tol == 3 or self.__Tol == 1:
            ax2.axvline(
                self.__mydict["LSL"], label="LSL", color="red", linestyle="--"
            )

        if self.__Tol == 3 or self.__Tol == 2:
            ax2.axvline(
                self.__mydict["HSL"], label="HSL", color="red", linestyle="--"
            )

        if self.__goal is True:
            ax2.axvline(
                self.__mydict["goal"],
                label="Goal",
                color="green",
                linestyle="--",
            )

        ax2.plot(
            np_values,
            stats.norm.pdf(np_values, dbl_Total_avg, SD),
            label="Short Term",
            color="grey",
            linestyle="--",
        )
        ax2.legend(loc="upper right")

        # ============================================================================================
        #                                Parameter
        # ============================================================================================
        ax3 = Fig_Cp.add_subplot(gs[0, 3])
        ax3.set_xticks([])
        ax3.set_yticks([])
        ax3.set_facecolor("white")
        ax3.set_ylim(0, 1)
        ax3.set_xlim(0, 1)

        # Short Term box
        str_msg = (
            f"{'':^25s}\n\n"
            f"{'':^2s}\n"
            f"{'':^2s}\n"
            f"{'':^2s}\n"
            f"{'':^2s}\n"
        )

        ax3.text(
            0.28,
            0.4,
            str_msg,
            bbox=dict(boxstyle="round", fc="w", color="black"),
            fontsize=12,
        )
        str_msg = f"{' ST Capability ':^13s}"
        ax3.text(0.3, 0.63, str_msg, fontsize=12)

        str_msg = (
            f"{'  Cp':8s}\n"
            f"{'  CpL':8s}\n"
            f"{'  CpU':8s}\n"
            f"{'  Cpk':8s}\n"
        )
        ax3.text(0.28, 0.4, str_msg, fontsize=12)

        str_msg = f"{Cp:<5} \n" f"{CpL:<5} \n" f"{CpU:<5} \n" f"{Cpk:<5} \n"
        ax3.text(0.5, 0.4, str_msg, fontsize=12)

        if Cpk > 1:
            str_adv = "\n   Process is capable"
            if self.__Tol == 3 and CpU > CpL:
                str_adv = str_adv + "\nand moved to the left"
            elif self.__Tol == 3 and CpU < CpL:
                str_adv = str_adv + "\nand moved to the rigth"

            str_adv = str_adv + f"\n\n     Cpk: {Cpk}>1\n"
            str_color = "mediumseagreen"
        else:
            str_adv = f"\nProcess is not capable\n\n     Cpk: {Cpk}<1\n"
            str_color = "red"

        ax3.annotate(
            str_adv,
            xy=(0.2, 0),
            bbox=dict(boxstyle="round", fc=str_color, color="black"),
            fontsize=12,
        )

        return Fig_Cp

    def __Pp_Report(self):  # noqa
        """Pp draws long term report

        Returns:
        ---------
        Fig_Lp : matplotlib figure
            long Term Report
        """
        # Determine avg and standard deviation for short and long term
        dbl_Param = 0
        dbl_SampleSize = 0
        dbl_iSampleSize = 0
        df_cp = self.__df_cp
        for item in df_cp["Batch"].unique():
            df_temp = df_cp[df_cp["Batch"] == item]
            dbl_iSampleSize = len(df_temp["Value"])
            dbl_SampleSize = dbl_SampleSize + dbl_iSampleSize
            dbl_Param = dbl_Param + (
                statistics.stdev(df_temp["Value"]) ** 2
            ) * (dbl_iSampleSize - 1)

        dbl_Total_avg = df_cp["Value"].mean()
        SDp = (
            dbl_Param / (dbl_SampleSize - len(df_cp["Batch"].unique()))
        ) ** (1 / 2)
        SD = statistics.stdev(df_cp["Value"])

        # Solve Capability parameters
        if self.__Tol == 3:
            Pp = round(
                (self.__mydict["HSL"] - self.__mydict["LSL"]) / (6 * SD), 2
            )
            PpL = round((dbl_Total_avg - self.__mydict["LSL"]) / (3 * SD), 2)
            PpU = round((self.__mydict["HSL"] - dbl_Total_avg) / (3 * SD), 2)
            Ppk = min(PpL, PpU)

            Cp = round(
                (self.__mydict["HSL"] - self.__mydict["LSL"]) / (6 * SDp), 2
            )
            CpL = round((dbl_Total_avg - self.__mydict["LSL"]) / (3 * SDp), 2)
            CpU = round((self.__mydict["HSL"] - dbl_Total_avg) / (3 * SDp), 2)
            Cpk = min(CpL, CpU)

        elif self.__Tol == 1:
            Pp = "*"
            PpL = round((dbl_Total_avg - self.__mydict["LSL"]) / (3 * SD), 2)
            PpU = "*"
            Ppk = PpL
            Cp = "*"
            CpL = round((dbl_Total_avg - self.__mydict["LSL"]) / (3 * SDp), 2)
            CpU = "*"
            Cpk = CpL
        elif self.__Tol == 2:
            Pp = "*"
            PpL = "*"
            PpU = round((self.__mydict["HSL"] - dbl_Total_avg) / (3 * SD), 2)
            Ppk = PpU
            Cp = "*"
            CpL = "*"
            CpU = round((self.__mydict["HSL"] - dbl_Total_avg) / (3 * SDp), 2)
            Cpk = CpU

        dbl_step = (df_cp["Value"].max() - df_cp["Value"].min()) * 0.01
        np_values = np.arange(
            df_cp["Value"].min(), df_cp["Value"].max(), dbl_step
        )

        Fig_Lp = plt.figure(figsize=(18, 12))
        Fig_Lp.set_facecolor("white")
        gs = mpl.gridspec.GridSpec(2, 4, wspace=0.2, hspace=0.2)
        Fig_Lp.suptitle(
            "\nProcess capability report for {}".format(
                self.__mydict["value"]
            ),
            fontsize=20,
        )

        # ============================================================================================
        #                                DATA
        # ============================================================================================
        ax1 = Fig_Lp.add_subplot(gs[0, 0])
        ax1.set_xticks([])
        ax1.set_yticks([])
        ax1.set_facecolor("white")
        ax1.set_ylim(0, 1)
        ax1.set_xlim(0, 1)

        str_msg = (
            f"{'':^40s}\n\n"
            f"{'':^35s}\n"
            f"{'':^35s}\n"
            f"{'':^35s}\n"
            f"{'':^35s}\n"
            f"{'':^35s}\n"
            f"{'':^35s}\n"
            f"{'':^35s}\n"
        )

        ax1.text(
            0.25,
            0.21,
            str_msg,
            bbox=dict(boxstyle="round", fc="w", color="black"),
            fontsize=12,
        )
        str_msg = f"{'Data Processing':^25s}"
        ax1.text(0.23, 0.58, str_msg, fontsize=12)

        str_msg = (
            f"{' LSL':15s}\n"
            f"{' Goal':15s}\n"
            f"{' HSL':15s}\n"
            f"{' Mean':15s}\n"
            f"{' Sample Size':15s}\n"
            f"{' Std. Dev (LT)':15s}\n"
            f"{' Std. Dev (ST)':15s}\n"
        )
        ax1.text(0.25, 0.2, str_msg, fontsize=12)

        str_msg = (
            f"{self.__mydict['LSL']:<10} \n"
            f"{self.__mydict['goal']:<10} \n"
            f"{self.__mydict['HSL']:<10} \n"
            f"{dbl_Total_avg:<10.3g} \n"
            f"{len(df_cp.index):<10.3g} \n"
            f"{SD:<10.3g} \n"
            f"{SDp:<10.3g} \n"
        )
        ax1.text(0.66, 0.2, str_msg, fontsize=12)

        # ============================================================================================
        #                                HISTOGRAM
        # ============================================================================================
        ax2 = Fig_Lp.add_subplot(gs[0, 1:3])
        ax2.hist(df_cp["Value"], bins=100, label="Data")
        ax2.set_yticks([])

        if self.__Tol == 3 or self.__Tol == 1:
            ax2.axvline(
                self.__mydict["LSL"], label="LSL", color="red", linestyle="--"
            )

        if self.__Tol == 3 or self.__Tol == 2:
            ax2.axvline(
                self.__mydict["HSL"], label="HSL", color="red", linestyle="--"
            )

        if self.__goal is True:
            ax2.axvline(
                self.__mydict["goal"],
                label="Goal",
                color="green",
                linestyle="--",
            )

        ax2.plot(
            np_values,
            stats.norm.pdf(np_values, dbl_Total_avg, SD),
            label="LT",
            color="black",
        )
        ax2.plot(
            np_values,
            stats.norm.pdf(np_values, dbl_Total_avg, SDp),
            label="ST",
            color="grey",
            linestyle="--",
        )
        ax2.legend(loc="upper right")

        # ============================================================================================
        #                                Parameter
        # ============================================================================================
        ax3 = Fig_Lp.add_subplot(gs[0, 3])
        ax3.set_xticks([])
        ax3.set_yticks([])
        ax3.set_facecolor("white")
        ax3.set_ylim(0, 1)
        ax3.set_xlim(0, 1)

        # Long Term box
        str_msg = (
            f"{'':^25s}\n\n"
            f"{'':^2s}\n"
            f"{'':^2s}\n"
            f"{'':^2s}\n"
            f"{'':^2s}\n"
        )

        ax3.text(
            0.02,
            0.4,
            str_msg,
            bbox=dict(boxstyle="round", fc="w", color="black"),
            fontsize=12,
        )
        str_msg = f"{' LT Capability ':^13s}"
        ax3.text(0.04, 0.63, str_msg, fontsize=12)

        str_msg = (
            f"{'  Pp':8s}\n"
            f"{'  PpL':8s}\n"
            f"{'  PpU':8s}\n"
            f"{'  Ppk':8s}\n"
        )
        ax3.text(0.02, 0.4, str_msg, fontsize=12)

        str_msg = f"{Pp:<5} \n" f"{PpL:<5} \n" f"{PpU:<5} \n" f"{Ppk:<5} \n"
        ax3.text(0.25, 0.4, str_msg, fontsize=12)

        # Short Term box
        str_msg = (
            f"{'':^25s}\n\n"
            f"{'':^2s}\n"
            f"{'':^2s}\n"
            f"{'':^2s}\n"
            f"{'':^2s}\n"
        )

        ax3.text(
            0.48,
            0.4,
            str_msg,
            bbox=dict(boxstyle="round", fc="w", color="black"),
            fontsize=12,
        )
        str_msg = f"{' ST Capability ':^13s}"
        ax3.text(0.5, 0.63, str_msg, fontsize=12)

        str_msg = (
            f"{'  Cp':8s}\n"
            f"{'  CpL':8s}\n"
            f"{'  CpU':8s}\n"
            f"{'  Cpk':8s}\n"
        )
        ax3.text(0.48, 0.4, str_msg, fontsize=12)

        str_msg = f"{Cp:<5} \n" f"{CpL:<5} \n" f"{CpU:<5} \n" f"{Cpk:<5} \n"
        ax3.text(0.7, 0.4, str_msg, fontsize=12)

        if Cpk > 1:
            str_adv = "\n   Process is capable"
            if self.__Tol == 3 and CpU > CpL:
                str_adv = str_adv + "\nand moved to the left"
            elif self.__Tol == 3 and CpU < CpL:
                str_adv = str_adv + "\nand moved to the rigth"

            str_adv = str_adv + f"\n\n     Cpk: {Cpk}>1\n"
            str_color = "mediumseagreen"
        else:
            str_adv = f"\nProcess is not capable\n\n     Cpk: {Cpk}<1\n"
            str_color = "red"

        ax3.annotate(
            str_adv,
            xy=(0.2, 0),
            bbox=dict(boxstyle="round", fc=str_color, color="black"),
            fontsize=12,
        )

        return Fig_Lp

    def Report(self):
        """Report is a figure that contain a chart where measures distirbution
        is showed with its tolerance. Then capability factor is determined
        and showed

        Returns:
        ---------
        myFigure : matplotlib figure
            Set of charts
        """
        if self.__AnalysisType == 1:
            myFigure = self.__Cp_Report()
        elif self.__AnalysisType == 2:
            myFigure = self.__Pp_Report()
        else:
            raise ValueError("Error Rep_01: Cp_Model is not defined correctly")

        return myFigure
